@@ -6,6 +6,7 @@ import os
 import platform
 import sys
 from collections import defaultdict
+from copy import deepcopy
 from datetime import datetime
 from pathlib import Path
 from typing import Any
@@ -20,7 +21,7 @@ from d2b.utils import prepend
 from d2b.utils import rsync
 from d2b.utils import splitext
 
-__version__ = "1.1.5"
+__version__ = "1.2.0"
 
 
 T = TypeVar("T")
@@ -29,9 +30,9 @@ T = TypeVar("T")
 class D2B:
     def __init__(
         self,
-        in_dirs: str | Path | list[str | Path],
+        in_dirs: str | Path | list[str] | list[Path],
         out_dir: str | Path,
-        config_file: Path,
+        config_file: str | Path,
         participant: str,
         session: str = defaults.session,
         options: dict[str, Any] = None,  # from the cli
@@ -40,6 +41,7 @@ class D2B:
         self.files: list[Path]  # set in self.run()
         self.descriptions: list[Description]  # set in self.run()
         self.matcher: Matcher  # set in self.run()
+        self.acquisitions: list[Acquisition]  # set in self.run()
 
         self.in_dirs = (
             [Path(d) for d in in_dirs] if isinstance(in_dirs, list) else [Path(in_dirs)]
@@ -53,37 +55,15 @@ class D2B:
 
         self._configure_logger()  # logging setup
 
-    @property
-    def in_dirs(self) -> list[Path]:
-        return self._in_dirs
-
-    @in_dirs.setter
-    def in_dirs(self, value: list[Path]):
-        dir_not_found = [d for d in value if not d.is_dir()]
-        if dir_not_found:
-            raise FileNotFoundError(dir_not_found)
-        self._in_dirs = value
-
     def load_config(self):
         self.config = pm.hook.load_config(  # type: ignore
             path=self.config_file,
             d2b=self,
         )
 
-    def pre_run_logs(self):
-        self.logger.info("--- d2b start ---")
-        self.logger.info("OS:version: %s", platform.platform())
-        self.logger.info("python:version: %s", sys.version.replace("\n", ""))
-        self.logger.info("d2b:version: %s", __version__)
-        self.logger.info("participant: %s", self.participant.bids_label)
-        self.logger.info("session: %s", self.participant.bids_session)
-        self.logger.info("config: %s", os.path.realpath(self.config_file))
-        self.logger.info("BIDS directory: %s", os.path.realpath(self.out_dir))
-
-        pm.hook.pre_run_logs(logger=self.logger, d2b=self)  # type: ignore
-
     def run(self):
-        self.pre_run_logs()
+        self._pre_run_logs()
+        self._check_in_dirs_exist()
 
         # make copies of the input directories
         dst_parent = self.d2b_dir / "src"
@@ -111,22 +91,26 @@ class D2B:
             Description.from_dict(i, d)
             for i, d in enumerate(self.config["descriptions"])
         ]
-        self.check_descriptions(self.descriptions)
+        self._check_for_effectively_nonunique_descriptions(self.descriptions)
 
         # run the matching algorithm
         self.matcher = Matcher(
+            self.files,
             self.participant,
             self.descriptions,
-            self.files,
             self.config,
             self.options,
         )
-        self.matcher.run()
+        unresolved_acquisitions = self.matcher.run()
+
+        # resolve IntendedFor Fields
+        resolver = IntendedForResolver(logger=self.logger)
+        self.acquisitions = resolver.resolve(unresolved_acquisitions)
 
         # run pre-move hooks
         self.logger.info("Running pre-move hooks")
         pm.hook.pre_move(  # type: ignore
-            acquisitions=self.matcher.acquisitions,
+            acquisitions=self.acquisitions,
             config=self.config,
             options=self.options,
             d2b=self,
@@ -134,11 +118,11 @@ class D2B:
 
         # move the files
         self.logger.info("Moving acquisitions into BIDS folder")
-        for acquisition in self.matcher.acquisitions:
+        for acquisition in self.acquisitions:
             pm.hook.move(  # type: ignore
                 out_dir=self.out_dir,
                 acquisition=acquisition,
-                acquisitions=self.matcher.acquisitions,
+                acquisitions=self.acquisitions,
                 config=self.config,
                 options=self.options,
                 d2b=self,
@@ -148,15 +132,40 @@ class D2B:
         self.logger.info("Running post-move hooks")
         pm.hook.post_move(  # type: ignore
             out_dir=self.out_dir,
-            acquisitions=self.matcher.acquisitions,
+            acquisitions=self.acquisitions,
             config=self.config,
             options=self.options,
             d2b=self,
         )
 
-    def check_descriptions(self, descriptions: list[Description]):
+    def _pre_run_logs(self):
+        self.logger.info("--- d2b start ---")
+        self.logger.info("OS:version: %s", platform.platform())
+        self.logger.info("python:version: %s", sys.version.replace("\n", ""))
+        self.logger.info("d2b:version: %s", __version__)
+        self.logger.info("participant: %s", self.participant.bids_label)
+        self.logger.info("session: %s", self.participant.bids_session)
+        self.logger.info("config: %s", os.path.realpath(self.config_file))
+        self.logger.info("BIDS directory: %s", os.path.realpath(self.out_dir))
+
+        pm.hook.pre_run_logs(logger=self.logger, d2b=self)  # type: ignore
+
+    def _check_in_dirs_exist(self):
+        dir_not_found = [d for d in self.in_dirs if not d.is_dir()]
+        if dir_not_found:
+            raise FileNotFoundError(dir_not_found)
+
+    def _check_for_effectively_nonunique_descriptions(
+        self,
+        descriptions: list[Description],
+    ):
         agg: dict[Description, list[Description]] = defaultdict(list)
         for d in descriptions:
+            # NOTE: this "non-uniqueness" detection procedure relies on
+            # the fact that desc1 == desc2 iff they have the same hash
+            # (since `d` is being used as a dict key), which is
+            # determined by hashing the tuple containing data_type,
+            # modality_label, and custom_labels
             agg[d].append(d)
 
         for ds in agg.values():
@@ -190,42 +199,44 @@ class D2B:
 class Matcher:
     def __init__(
         self,
+        files: list[Path],
         participant: Participant,
         descriptions: list[Description],
-        files: list[Path],
         config: dict[str, Any] = None,  # d2b config
         options: dict[str, Any] = None,  # from the cli
+        logger: logging.Logger = None,
     ):
         self.participant = participant
         self.descriptions = descriptions
         self.files = files
         self.config = config or {}
         self.options = options or {}
-
-        self.logger = logging.getLogger(__name__)
+        self.logger = logger or logging.getLogger(__name__)
 
         # populated in self.find_matches()
         self.file_to_acq: dict[Path, list[Acquisition]] = {fp: [] for fp in files}
-        # populated in self.remove_multimatch()
+        # populated in self.filter_unique_matches()
         self.acquisitions: list[Acquisition] = []
 
-    def run(self):
+    def run(self) -> list[Acquisition]:
         self.find_matches()
         self.filter_unique_matches()
         self.dedup_runs()
-        self.resolve_matches_intended_for()
+        return self.acquisitions
 
     def find_matches(self):
         possible_matches = itertools.product(self.files, self.descriptions)
         for fp, description in possible_matches:
             criteria = description.data.get("criteria")
+            if criteria is None:
+                continue
             possible_link = pm.hook.is_link(  # type: ignore
                 path=fp,
                 criteria=criteria,
                 config=self.config,
                 options=self.options,
             )
-            if criteria is None or not any(possible_link):
+            if not any(possible_link):
                 continue
             acquisition = Acquisition(fp, self.participant, description.copy())
             self.file_to_acq[fp].append(acquisition)
@@ -293,55 +304,121 @@ class Matcher:
             if len(positions) > 1:
                 yield key, positions
 
-    def resolve_matches_intended_for(self):
+
+class IntendedForResolver:
+    """Object to handle resolving `IntendedFor` fields among a list of acquisitions
+
+    Note:
+        Calling `IntendedForResolver.resolve()` on a sequence of acquisitions
+        will mutate the underlying acquisitions, specifically, only those
+        acquisitions for which `acq.description.intended_for is not None`
+    """
+
+    def __init__(self, logger: logging.Logger | None = None):
+        self.logger = logger or logging.getLogger(__name__)
+
+        # set in self.resolve()
+        self.acquisitions: list[Acquisition]
+
+    def resolve(  # noqa: C901
+        self,
+        acquisitions: list[Acquisition] = None,
+    ) -> list[Acquisition]:
+        self.acquisitions = acquisitions or []
         for acq in self.acquisitions:
             description = acq.description
             intended_for = description.intended_for
 
             if intended_for is None:
+                # this description does not contain an IntendedFor property
                 continue
 
             elif isinstance(intended_for, int) or isinstance(intended_for, str):
-                fp = self._resolve_intended_for_path(acq, intended_for)
-                if fp is None:
+                # IntendedFor is a single item (int | str)
+                fps = self._resolve_intended_for_paths(acq, intended_for)
+                if fps is None:
+                    # the target acquisition was not found
                     continue
-                acq.data["IntendedFor"] = str(fp)
+
+                if isinstance(fps, Path):
+                    # one target acquisition was found
+                    acq.data["IntendedFor"] = str(fps)
+                elif isinstance(fps, list):
+                    # multiple target acqs were found for this item
+                    acq.data["IntendedFor"] = [str(fp) for fp in fps]
+                else:
+                    # something went wrong ...
+                    raise TypeError(
+                        "Expected resolved IntendedFor paths to be Path | "
+                        f"Path[] | None. Found [{fps!r}]",
+                    )
 
             elif isinstance(intended_for, list):
+                # IntendedFor is a list
                 for _intended_for in intended_for:
-                    fp = self._resolve_intended_for_path(acq, _intended_for)
-                    if fp is None:
+                    fps = self._resolve_intended_for_paths(acq, _intended_for)
+                    if fps is None:
+                        # none of the target acquisitions were found
                         continue
                     if acq.data.get("IntendedFor") is None:
+                        # IntendedFor has not yet been initialized
                         acq.data["IntendedFor"] = []
-                    acq.data["IntendedFor"].append(str(fp))
 
+                    if isinstance(fps, Path):
+                        # one target acq was found for this item
+                        acq.data["IntendedFor"].append(str(fps))
+                    elif isinstance(fps, list):
+                        # multiple target acqs were found for this item
+                        if isinstance(fps, Path):
+                            acq.data["IntendedFor"].extend([str(fp) for fp in fps])
+                    else:
+                        # something went wrong ...
+                        raise TypeError(
+                            "Expected resolved IntendedFor paths to be Path | "
+                            f"Path[] | None. Found [{fps!r}]",
+                        )
             else:
                 raise ValueError(
                     f"Invalid IntendedFor value [{intended_for}] in "
                     f"description at index [{description.index}]",
                 )
 
-    def _resolve_intended_for_path(
+        return self.acquisitions
+
+    def _resolve_intended_for_paths(
         self,
         acq: Acquisition,
         intended_for: int | str,
-    ) -> Path | None:
-        target_acq = self._resolve_intended_for(self.acquisitions, intended_for)
-        if target_acq is None:
+    ) -> Path | list[Path] | None:
+        target_acqs = self._resolve_intended_for(self.acquisitions, intended_for)
+        if target_acqs is None:
             return
-        ext = associated_nii_ext(target_acq.src_file)
-        if ext is None:
-            self._log_associated_nii_not_found(acq, target_acq)
-            return
-        bids_path = target_acq.dst_root.with_suffix(ext)
-        return bids_path.relative_to(target_acq.participant.subject_directory)
+        if len(target_acqs) == 1:
+            target_acq = target_acqs[0]
+            ext = associated_nii_ext(target_acq.src_file)
+            if ext is None:
+                self._log_associated_nii_not_found(acq, target_acq)
+                return
+            bids_path = target_acq.dst_root.with_suffix(ext)
+            return bids_path.relative_to(target_acq.participant.subject_directory)
+        else:
+            paths: list[Path] = []
+            for target_acq in target_acqs:
+                ext = associated_nii_ext(target_acq.src_file)
+                if ext is None:
+                    self._log_associated_nii_not_found(acq, target_acq)
+                    return
+                bids_path = target_acq.dst_root.with_suffix(ext)
+                paths.append(
+                    bids_path.relative_to(target_acq.participant.subject_directory),
+                )
+            return paths
 
     @staticmethod
     def _resolve_intended_for(
         acquisitions: list[Acquisition],
         intended_for: int | str,
-    ) -> Acquisition | None:
+    ) -> list[Acquisition] | None:
         if isinstance(intended_for, int):
             match_refs = [
                 a for a in acquisitions if a.description.index == intended_for
@@ -364,7 +441,7 @@ class Matcher:
             # skip adding this IntendedFor to the sidecar's data.
             return
 
-        return match_refs[0]
+        return match_refs
 
     def _log_associated_nii_not_found(self, acq: Acquisition, target_acq: Acquisition):
         msg = (
@@ -402,11 +479,7 @@ class Acquisition:
         self.data = data or {}
 
     def __eq__(self, o):
-        return (
-            self.participant.prefix == o.participant.prefix
-            and self.description.data_type == o.description.data_type
-            and self.description.suffix == o.description.suffix
-        )
+        return self.dst_root == o.dst_root
 
     @property
     def src_root(self) -> Path:
@@ -437,6 +510,12 @@ class Participant:
     def __repr__(self):
         return f"{self.__class__.__name__}({self.label!r}, {self.session!r})"
 
+    def __eq__(self, o):
+        return self.label == o.label and self.session == o.session
+
+    def __hash__(self):
+        return hash((self.label, self.session))
+
     # getters/setters
     @property
     def label(self):
@@ -518,14 +597,15 @@ class Description:
 
     @classmethod
     def from_dict(cls, index: int, data: dict[str, Any]):
+        _d = deepcopy(data)
         return cls(
             index=index,
-            data_type=data["dataType"],
-            modality_label=data["modalityLabel"],
-            custom_labels=data.get("customLabels", ""),
-            sidecar_changes=data.get("sidecarChanges"),
-            intended_for=data.get("IntendedFor"),
-            data=data,
+            data_type=_d.pop("dataType"),
+            modality_label=_d.pop("modalityLabel"),
+            custom_labels=_d.pop("customLabels", ""),
+            sidecar_changes=_d.pop("sidecarChanges", None),
+            intended_for=_d.pop("IntendedFor", None),
+            data=_d,
         )
 
     def copy(self):
@@ -534,9 +614,9 @@ class Description:
             data_type=self.data_type,
             modality_label=self.modality_label,
             custom_labels=self.custom_labels,
-            sidecar_changes=self.sidecar_changes,
-            intended_for=self.intended_for,
-            data=self.data,
+            sidecar_changes=deepcopy(self.sidecar_changes),
+            intended_for=deepcopy(self.intended_for),
+            data=deepcopy(self.data),
         )
 
     # getter/setters
@@ -636,15 +716,16 @@ class FilenameEntities:
     @classmethod
     def parse(cls, entities: str | dict[str, str]) -> dict[str, str]:
         if isinstance(entities, dict):
-            return entities
+            return {k: cls._sanitize_entity_value(v) for k, v in entities.items()}
+
         stripped_entities = entities.strip().strip("_")
         if stripped_entities == "":
             return {}
         entity_strs = entities.strip().strip("_").split("_")
         return dict(map(cls._split_entity_str, entity_strs))
 
-    @staticmethod
-    def _split_entity_str(s: str) -> tuple[str, str]:
+    @classmethod
+    def _split_entity_str(cls, s: str) -> tuple[str, str]:
         parts = s.strip().split("-")
         if len(parts) < 2:
             m = (
@@ -653,9 +734,13 @@ class FilenameEntities:
             )
             raise ValueError(m)
         k = parts[0]
-        v = "".join(filter(str.isalnum, "".join(parts[1:])))
+        v = cls._sanitize_entity_value("".join(parts[1:]))
         return k, v
 
+    @staticmethod
+    def _sanitize_entity_value(v: str):
+        return "".join(filter(str.isalnum, v))
+
 
 def _setup_logging(log_level, logFile=None):
     """Setup logging configuration"""
